@@ -16,7 +16,7 @@
  *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307,
  *  USA.
  *  
- *  $Id: ydpdict.c,v 1.6 2007-07-30 16:10:12 wojtekka Exp $
+ *  $Id: ydpdict.c 52 2007-12-16 09:49:42Z wojtekka $
  */
 
 #include <sys/types.h>
@@ -31,6 +31,32 @@
 #include "config.h"
 #endif
 
+/**
+ * Word array entry.
+ */
+typedef struct {
+	char *word;			/** Word in output encoding */
+	uint32_t index;			/** Word definition index in database */
+} ydpdict_word_t;
+
+/**
+ * Dictionary structure.
+ */
+typedef struct {
+	FILE *dat;			/** Database file handle */
+	FILE *idx;			/** Index file handle */
+
+	int count;			/** Word count */
+	ydpdict_word_t *words;		/** Word array */
+	
+	ydpdict_encoding_t encoding;	/** Output encoding */
+ 
+	int xhtml_header;		/** XHTML header output flag */
+	char *xhtml_title;		/** XHTML title */
+	char *xhtml_style;		/** XHTML custom style */
+	int xhtml_use_style;		/** XHTML style usage flag */
+} ydpdict_priv_t;
+
 #define ATTR_B 1
 #define ATTR_CF0 2
 #define ATTR_CF1 4
@@ -41,6 +67,7 @@
 #define ATTR_F1 128
 #define ATTR_I 256
 #define ATTR_CF5 1024
+#define ATTR_SA 2048
 
 /**
  * \brief Conversion table from phonetic characters to UTF-8
@@ -70,7 +97,7 @@ static char *ydpdict_windows1250_to_utf8_table[128] =
 	"Č", "É", "Ę", "Ë", "Ě", "Í", "Î", "Ď", 
 	"Đ", "Ń", "Ň", "Ó", "Ô", "Ő", "Ö", "×", 
 	"Ř", "Ů", "Ú", "Ű", "Ü", "Ý", "Ţ", "ß", 
-	"ŕ", "á", "â", "ă", "ä", "ĺ", "ć", "ç", 
+	"à", "á", "â", "ă", "ä", "ĺ", "ć", "ç", 
 	"č", "é", "ę", "ë", "ě", "í", "î", "ï", 
 	"đ", "ń", "ň", "ó", "ô", "ő", "ö", "÷", 
 	"ř", "ů", "ú", "ű", "ü", "ý", "ţ", "˙", 
@@ -113,34 +140,30 @@ static inline uint16_t ydpdict_fix16(uint16_t value)
 }
 
 /**
- * \brief Open dictionary and read words' indices
+ * \brief Open dictionary and read index
  * 
  * The common mistake is to supply lowercase names, while the files have
  * uppercase names.
  * 
- * \param dict allocated dictionary description
  * \param dat data file path
  * \param idx index file path
  * \param encoding output encoding for XHTML
  * 
- * \return 0 on success, -1 on error
+ * \return Pointer to allocated structure or NULL on error
  */
-int ydpdict_open(ydpdict_t *dict, const char *dat, const char *idx, ydpdict_encoding_t encoding)
+ydpdict_t *ydpdict_open(const char *dat, const char *idx, ydpdict_encoding_t encoding)
 {
+	ydpdict_priv_t *dict = NULL;
 	uint32_t index;
+	uint16_t count;
 	int i, j;
 
-	/* Clear ydpdict_t and set defaults */
+	if (!(dict = calloc(1, sizeof(ydpdict_priv_t))))
+		goto failure;
+
+	/* Set defaults */
 
-	dict->idx = NULL;
-	dict->dat = NULL;
-	dict->word_count = 0;
-	dict->words = NULL;
-	dict->indices = NULL;
 	dict->xhtml_header = 1;
-	dict->xhtml_title = NULL;
-	dict->xhtml_style = NULL;
-	dict->xhtml_use_style = 0;
 	dict->encoding = encoding;
 		
 	/* Open files */
@@ -156,23 +179,16 @@ int ydpdict_open(ydpdict_t *dict, const char *dat, const char *idx, ydpdict_enco
 	if (fseek(dict->idx, 8, SEEK_SET) == (off_t) -1)
 		goto failure;
 
-	dict->word_count = 0;
-	
-	if (fread(&dict->word_count, sizeof(dict->word_count), 1, dict->idx) != 1)
+	if (fread(&count, sizeof(count), 1, dict->idx) != 1)
 		goto failure;
 	
-	dict->word_count = ydpdict_fix16(dict->word_count);
+	dict->count = ydpdict_fix16(count);
 
 	/* Allocate memory */
 
-	if (!(dict->indices = calloc(dict->word_count, sizeof(uint32_t))))
+	if (!(dict->words = calloc(dict->count, sizeof(ydpdict_word_t))))
 		goto failure;
 	
-	if (!(dict->words = calloc(dict->word_count + 1, sizeof(char*))))
-		goto failure;
-
-	dict->words[dict->word_count] = NULL;
-    
 	/* Read index table offset */
 
 	if (fseek(dict->idx, 16, SEEK_SET) == (off_t) -1)
@@ -198,12 +214,10 @@ int ydpdict_open(ydpdict_t *dict, const char *dat, const char *idx, ydpdict_enco
 		if (fseek(dict->idx, 4, SEEK_CUR) == (off_t) -1)
 			goto failure;
 
-		dict->indices[i] = 0;
-		
-		if (fread(&dict->indices[i], 4, 1, dict->idx) != 1)
+		if (fread(&index, sizeof(index), 1, dict->idx) != 1)
 			goto failure;
 		
-		dict->indices[i] = ydpdict_fix32(dict->indices[i]);
+		dict->words[i].index = ydpdict_fix32(index);
 
 		j = 0;
 
@@ -229,17 +243,18 @@ int ydpdict_open(ydpdict_t *dict, const char *dat, const char *idx, ydpdict_enco
 			}
 		} while (j < sizeof(buf) && buf[j++]);
 
-		if (!(dict->words[i] = strdup(buf)))
+		if (!(dict->words[i].word = strdup(buf)))
 			goto failure;
 
-	} while (++i < dict->word_count);
+	} while (++i < dict->count);
 
-	return 0;
+	return dict;
 	
 failure:
-	ydpdict_close(dict);
+	if (dict)
+		ydpdict_close(dict);
 
-	return -1;
+	return NULL;
 }
 
 /**
@@ -251,20 +266,21 @@ failure:
  *
  * \return allocated buffer with definition on success, NULL on error
  */
-char *ydpdict_read_rtf(const ydpdict_t *dict, uint32_t def)
+char *ydpdict_read_rtf(const ydpdict_t *pdict, int def)
 {
+	const ydpdict_priv_t *dict = pdict;
 	char *text = NULL;
 	uint32_t len = 0;
 
-	if (!dict || def >= dict->word_count) {
+	if (!dict || def >= dict->count) {
 		errno = EINVAL;
 		return NULL;
 	}
 
-	if (fseek(dict->dat, dict->indices[def], SEEK_SET) == (off_t) -1)
+	if (fseek(dict->dat, dict->words[def].index, SEEK_SET) == (off_t) -1)
 		goto failure;
 
-	if (fread(&len, 4, 1, dict->dat) != 1)
+	if (fread(&len, sizeof(len), 1, dict->dat) != 1)
 		goto failure;
 	
 	len = ydpdict_fix32(len);
@@ -292,20 +308,57 @@ failure:
  * \param dict dictionary description
  * \param word complete or partial word
  * 
- * \return definition index on success, (uint32_t) -1 on error
+ * \return definition index on success, -1 on error
  */
-uint32_t ydpdict_find(const ydpdict_t *dict, const char *word)
+int ydpdict_find_word(const ydpdict_t *pdict, const char *word)
 {
+	const ydpdict_priv_t *dict = pdict;
 	int i = 0;
 
 	if (!dict)
-		return (uint32_t) -1;
+		return -1;
 	
-	for (; i < dict->word_count; i++)
-		if (!strncasecmp(dict->words[i], word, strlen(word)))
+	for (; i < dict->count; i++) {
+		if (!strncasecmp(dict->words[i].word, word, strlen(word)))
 			return i;
+	}
 	
-	return (uint32_t) -1;
+	return -1;
+}
+
+/**
+ * \brief Returns number of words in dictionary
+ *
+ * \param dict dictionary description
+ * 
+ * \return word count on success, -1 on error
+ */
+int ydpdict_get_count(const ydpdict_t *pdict)
+{
+	const ydpdict_priv_t *dict = pdict;
+
+	if (!dict)
+		return -1;
+
+	return dict->count;
+}
+
+/**
+ * \brief Read word from dictionary
+ *
+ * \param dict dictionary description
+ * \param def word index
+ *
+ * \return constant buffer with word on success, NULL on error
+ */
+const char *ydpdict_get_word(const ydpdict_t *pdict, int def)
+{
+	const ydpdict_priv_t *dict = pdict;
+
+	if (!dict || def >= dict->count)
+		return NULL;
+
+	return dict->words[def].word;
 }
 
 /**
@@ -315,20 +368,18 @@ uint32_t ydpdict_find(const ydpdict_t *dict, const char *word)
  *
  * \return 0 on success, -1 on error
  */
-int ydpdict_close(ydpdict_t *dict)
+int ydpdict_close(ydpdict_t *pdict)
 {
-	if (dict->indices) {
-		free(dict->indices);
-		dict->indices = NULL;
-	}
-	
+	ydpdict_priv_t *dict = pdict;
+
+	if (!dict)
+		return -1;
+
 	if (dict->words) {
 		int i = 0;
 
-		while (dict->words[i]) {
-			free(dict->words[i]);
-			i++;
-		}
+		for (i = 0; i < dict->count; i++)
+			free(dict->words[i].word);
 		
 		free(dict->words);
 		dict->words = NULL;
@@ -354,6 +405,8 @@ int ydpdict_close(ydpdict_t *dict)
 		dict->xhtml_style = NULL;
 	}
 
+	free(dict);
+
 	return 0;
 }
 
@@ -399,10 +452,11 @@ static int ydpdict_append(char **buf, int *len, const char *str)
  *
  * \return allocated buffer with definition on success, NULL on error
  */
-char *ydpdict_read_xhtml(const ydpdict_t *dict, uint32_t def)
+char *ydpdict_read_xhtml(const ydpdict_t *pdict, int def)
 {
+	const ydpdict_priv_t *dict = pdict;
 	char *buf = NULL;
-	int attr_stack[16], level = 0, attr = 0, block_begin = 0;
+	int attr_stack[16], block_stack[16], level = 0, attr = 0, block_begin = 0;
 	int paragraph = 1, margin = 0, buf_len;
 	unsigned char *rtf, *rtf_orig;
 
@@ -441,21 +495,22 @@ char *ydpdict_read_xhtml(const ydpdict_t *dict, uint32_t def)
 		
 		APPEND("<?xml version=\"1.0\"?>\n<!DOCTYPE html PUBLIC \"-//W3C//DTD XHTML 1.1//EN\" \"http://www.w3.org/TR/xhtml11/DTD/xhtml11.dtd\">\n<html xmlns=\"http://www.w3.org/1999/xhtml\"><head>");
 
-		if (dict->xhtml_title) {
-			APPEND("<title>");
-			APPEND(dict->xhtml_title);
-			APPEND("</title>");
-		}
+		APPEND("<title>");
+		APPEND((dict->xhtml_title) ? dict->xhtml_title : "");
+		APPEND("</title>");
+
 		if (charset) {
 			APPEND("<meta http-equiv=\"Content-type\" content=\"text/html; charset=");
 			APPEND(charset);
 			APPEND("\" />");
 		}
+
 		if (dict->xhtml_style) {
 			APPEND("<style>");
 			APPEND(dict->xhtml_style);
 			APPEND("</style>");
 		}
+
 		APPEND("</head><body>");
 	}
 
@@ -464,11 +519,19 @@ char *ydpdict_read_xhtml(const ydpdict_t *dict, uint32_t def)
 	while (*rtf) {
 		switch (*rtf) {
 			case '{':
-				if (level < 16)
-					attr_stack[level++] = attr;
+				if (level < 16) {
+					attr_stack[level] = attr;
+					block_stack[level] = block_begin;
+					level++;
+				}
+
 				block_begin = 1;
+
 				attr = 0;
 				
+				if (margin && !(attr_stack[level - 1] & ATTR_SA))
+					attr |= ATTR_SA;
+				
 				rtf++;
 				break;
 
@@ -476,21 +539,32 @@ char *ydpdict_read_xhtml(const ydpdict_t *dict, uint32_t def)
 				if (!level)
 					break;
 
-				if ((attr & ATTR_SUPER))
-					APPEND("</sup>");
+				// don't end tags if they haven't started.
+				//
+				if (!block_begin) {
+					if ((attr & ATTR_SUPER))
+						APPEND("</sup>");
 
-				if ((attr & (ATTR_CF0 | ATTR_CF1 | ATTR_CF2 | ATTR_CF5)))
-					APPEND("</span>");
+					if ((attr & (ATTR_CF0 | ATTR_CF1 | ATTR_CF2 | ATTR_CF5)))
+						APPEND("</span>");
 				
-				if ((attr & ATTR_I))
-					APPEND("</i>");
+					if ((attr & ATTR_I))
+						APPEND("</i>");
 
-				if ((attr & ATTR_B))
-					APPEND("</b>");
+					if ((attr & ATTR_B))
+						APPEND("</b>");
+				}
 
+				if (margin && (attr & ATTR_SA)) {
+					APPEND("</div><p>");
+					margin = 0;
+				}
+				
 				paragraph = 0;
 
-				attr = attr_stack[--level];
+				level--;
+				attr = attr_stack[level];
+				block_begin = block_stack[level];
 
 				rtf++;
 				break;
@@ -515,11 +589,20 @@ char *ydpdict_read_xhtml(const ydpdict_t *dict, uint32_t def)
 				token[len] = 0;
 
 				if (!strcmp(token, "par") && margin) {
-					APPEND("</span>");
+					APPEND("</div><p>");
+					attr &= ~ATTR_SA;
 					margin = 0;
 				}
 
 				if (!strcmp(token, "pard")) {
+					// workaround for broken RTF in "żreć",
+					// where \pard is inside \i block
+					//
+					if (level > 0 && (attr & ATTR_I)) {
+						APPEND("</i>");
+						attr &= ~ATTR_I;
+					}
+
 					APPEND("</p><p>");
 					paragraph = 1;
 				}
@@ -527,29 +610,33 @@ char *ydpdict_read_xhtml(const ydpdict_t *dict, uint32_t def)
 				if (!strcmp(token, "line") && !paragraph)
 					APPEND("<br />");
 
-				if (!strcmp(token, "b"))
-					attr |= ATTR_B;
+				if (block_begin) {
+					if (!strcmp(token, "b"))
+						attr |= ATTR_B;
 
-				if (!strcmp(token, "i"))
-					attr |= ATTR_I;
+					if (!strcmp(token, "i"))
+						attr |= ATTR_I;
 
-				if (!strcmp(token, "cf0"))
-					attr |= ATTR_CF0;
+					if (!strcmp(token, "cf0"))
+						attr |= ATTR_CF0;
 
-				if (!strcmp(token, "cf1"))
-					attr |= ATTR_CF1;
+					if (!strcmp(token, "cf1"))
+						attr |= ATTR_CF1;
 
-				if (!strcmp(token, "cf2"))
-					attr |= ATTR_CF2;
+					if (!strcmp(token, "cf2"))
+						attr |= ATTR_CF2;
 	
-				if (!strcmp(token, "cf5"))
-					attr |= ATTR_CF5;
+					if (!strcmp(token, "cf5"))
+						attr |= ATTR_CF5;
+				}
 
 				if (!strcmp(token, "qc"))
 					attr |= ATTR_QC;
 
-				if (!strcmp(token, "super"))
+				if (!strcmp(token, "super")) {
 					attr |= ATTR_SUPER;
+					block_begin = 1;
+				}
 
 				if (!strncmp(token, "f", 1) && strcmp(token, "f1"))
 					attr |= ATTR_F;
@@ -559,7 +646,10 @@ char *ydpdict_read_xhtml(const ydpdict_t *dict, uint32_t def)
 
 				if (!strncmp(token, "sa", 2)) {
 					if (!margin) {
-						APPEND("<span style=\"display: block; margin-left: 1em;\">");
+						if (dict->xhtml_use_style)
+							APPEND("</p><div class=\"example\">");
+						else
+							APPEND("</p><div style=\"margin: -1em 2em auto 2em;\">");
 						margin = 1;
 					} else
 						APPEND("<br />");
@@ -569,7 +659,7 @@ char *ydpdict_read_xhtml(const ydpdict_t *dict, uint32_t def)
 			}
 
 			default:
-				if (block_begin) {
+				if (block_begin && *rtf != ' ') {
 					block_begin = 0;
 
 					if ((attr & ATTR_B))
@@ -616,6 +706,14 @@ char *ydpdict_read_xhtml(const ydpdict_t *dict, uint32_t def)
 						APPEND("<sup>");
 				}
 
+				// workaround for square bracket like in
+				// {[\f1\cf5 pronunciation]}
+				//
+				if (*rtf == ']' && (attr & ATTR_F1) && (attr & ATTR_CF5)) {
+					attr &= ~ATTR_CF5;
+					APPEND("</span>");
+				}
+
 				if (!(attr & ATTR_QC) && (dict->encoding == YDPDICT_ENCODING_UTF8)) {
 					if ((attr & ATTR_F1) && *rtf > 127 && *rtf < 160)
 						APPEND(ydpdict_phonetic_to_utf8_table[*rtf - 128]);
@@ -623,6 +721,12 @@ char *ydpdict_read_xhtml(const ydpdict_t *dict, uint32_t def)
 						APPEND(ydpdict_windows1250_to_utf8_table[*rtf - 128]);
 					else if (*rtf == 127)
 						APPEND("~");
+					else if (*rtf == '&')
+						APPEND("&amp;");
+					else if (*rtf == '<')
+						APPEND("&lt;");
+					else if (*rtf == '>')
+						APPEND("&gt;");
 					else {
 						char tmp[2] = { *rtf, 0 };
 						APPEND(tmp);
@@ -669,8 +773,10 @@ failure:
  *
  * \result 0 on success, -1 on error
  */
-int ydpdict_xhtml_set_style(ydpdict_t *dict, const char *style)
+int ydpdict_set_xhtml_style(ydpdict_t *pdict, const char *style)
 {
+	ydpdict_priv_t *dict = pdict;
+
 	if (!dict) {
 		errno = EINVAL;
 		return -1;
@@ -695,8 +801,10 @@ int ydpdict_xhtml_set_style(ydpdict_t *dict, const char *style)
  *
  * \result 0 on success, -1 on error
  */
-int ydpdict_xhtml_set_title(ydpdict_t *dict, const char *title)
+int ydpdict_set_xhtml_title(ydpdict_t *pdict, const char *title)
 {
+	ydpdict_priv_t *dict = pdict;
+
 	if (!dict) {
 		errno = EINVAL;
 		return -1;
@@ -721,8 +829,10 @@ int ydpdict_xhtml_set_title(ydpdict_t *dict, const char *title)
  *
  * \result 0 on success, -1 on error
  */
-int ydpdict_xhtml_set_header(ydpdict_t *dict, int header)
+int ydpdict_set_xhtml_header(ydpdict_t *pdict, int header)
 {
+	ydpdict_priv_t *dict = pdict;
+
 	if (!dict) {
 		errno = EINVAL;
 		return -1;
@@ -734,15 +844,21 @@ int ydpdict_xhtml_set_header(ydpdict_t *dict, int header)
 }
 
 /**
- * \brief Toggle XHTML style usage
+ * \brief Toggle CSS style in XHTML
+ *
+ * If CSS style is enabled, all text attributes are described by CSS classes.
+ * Otherwise all style information is embedded directly in style tag
+ * atttributes.
  *
  * \param dict dictionary description
  * \param use_style style usage flag
  *
  * \result 0 on success, -1 on error
  */
-int ydpdict_xhtml_set_use_style(ydpdict_t *dict, int use_style)
+int ydpdict_set_xhtml_use_style(ydpdict_t *pdict, int use_style)
 {
+	ydpdict_priv_t *dict = pdict;
+
 	if (!dict) {
 		errno = EINVAL;
 		return -1;
